import base64


class Base64Util:
    @staticmethod
    def encode(input_string: str) -> str:
        string_bytes = input_string.encode('utf-8')
        encoded_bytes = base64.b64encode(string_bytes)
        return encoded_bytes.decode('utf-8')

    @staticmethod
    def decode(encoded_string: str) -> str:
        encoded_bytes = encoded_string.encode('utf-8')
        decoded_bytes = base64.b64decode(encoded_bytes)
        return decoded_bytes.decode('utf-8')



from typing import Optional
from pydantic import BaseModel

from nkunyim_util.models.location_model import LocationModel
from nkunyim_util.models.user_agent_model import UserAgentModel


class LoggingModel(BaseModel):
    xan: str
    service_name: str
    location: Optional[LocationModel]
    user_agent: Optional[UserAgentModel]


from django.core.management.base import BaseCommand, CommandError

from nkunyim_util.management.commands.commands import (
    DataxNewCommand, 
    DataxRevCommand, 
    DataxRunCommand
)



class Command(BaseCommand):
    help = 'Send email reminders to inactive users'

    def add_arguments(self, parser):
        parser.add_argument(
            '--new',
            type=str,
            help='Generate a new DataxMigration file with the given name.'
        )

        parser.add_argument(
            '--run',
            type=str,
            help='Run a DataxMigration from a give file name.'
        )

        parser.add_argument(
            '--rev',
            type=str,
            help='Reverse a DataxMigration from a given file name.'
        )
        
    def handle(self, *args, **options):
        self.stdout.write(self.style.HTTP_INFO("Datax Migration ") + self.style.SQL_KEYWORD("Started!"))
        
        if options['new']:
            info = "Generating a new Datax Migration file: "
            msg = "A new Datax Migration file has been generated successfully at: "
            cmd = DataxNewCommand(input_name=options['new'])
        elif options['run']:
            info = "Running Datax Migration from the file: "
            msg = "Datax Migration has been migrated from the file: "
            cmd = DataxRunCommand(input_name=options['run'])
        elif options['rev']:
            info = "Reversing Datax Migration from the file: "
            msg = "Datax Migration has been reversed from the file: "
            cmd = DataxRevCommand(input_name=options['rev'])
        else:
            raise CommandError('The --some-required-option is missing.')
            
            
        self.stdout.write(self.style.HTTP_SUCCESS(info) + self.style.HTTP_INFO(cmd.python_file_name))
        
        res = cmd.go()
        if res: 
            self.stdout.write(self.style.HTTP_SUCCESS(res))
            self.stdout.write(self.style.HTTP_INFO("Datax Migration ") + self.style.ERROR_OUTPUT("Failed!"))
        else:
            self.stdout.write(self.style.HTTP_SUCCESS(msg) + self.style.HTTP_INFO(cmd.pyhton_file_path))
            self.stdout.write(
                self.style.HTTP_SUCCESS("To run the migration file, use: ") + 
                self.style.HTTP_INFO("python manage.py datax --run " + cmd.file_name)
            )
            self.stdout.write(self.style.HTTP_INFO("Datax Migration ") + self.style.SUCCESS("Successful!"))



from django.http import HttpRequest

from nkunyim_util.models.application_model import AppModel
from nkunyim_util.models.context_model import ContextModel
from nkunyim_util.services.account_service import AccountService
from nkunyim_util.services.application_service import ApplicationService
from nkunyim_util.services.location_service import LocationService
from nkunyim_util.services.nation_service import NationService
from nkunyim_util.services.page_service import PageService
from nkunyim_util.services.user_agent_service import UserAgentService
from nkunyim_util.services.user_service import UserService


class ContextService:
    
    def __init__(self, req: HttpRequest) -> None:
        self.req = req
        
    
    def create(self) -> ContextModel:
        page_model = PageService(req=self.req)
        page_data = page_model.get()
        
        application_model = ApplicationService(req=self.req, session_key=page_data.root)
        application_data = application_model.get()
        
        app_data = AppModel(**application_data.__dict__)
        
        user_model = UserService(req=self.req)
        user_data = user_model.app()

        account_model = AccountService(req=self.req, session_key=page_data.root, application_id=application_data.id)
        account_data = account_model.get()

        location_model = LocationService(req=self.req, session_key=page_data.root)
        location_data = location_model.get()
        
        nation_model = NationService(req=self.req, session_key=page_data.root, code=location_data.country_code)
        nation_data = nation_model.get()
        
        user_agent_model = UserAgentService(req=self.req, session_key=page_data.root)
        user_agent_data = user_agent_model.get()
        
        data = {
            **app_data.__dict__,
            'page': page_data.__dict__,
            'location': location_data.__dict__,
            'user_agent': user_agent_data.__dict__,
            'navs': account_data.navs if account_data.navs else None,
            'uix': account_data.uix,
            'nation': nation_data.__dict__,
            'user': user_data.__dict__ if user_data else None,
            'root': page_data.root,
        }
        
        return ContextModel(**data)



import base64, json
import requests
from django.conf import settings


LOG_TYPE_GEN = "GEN"
LOG_TYPE_API = "API"
LOG_TYPE_IAM = "IAM"

LOG_TYPE_KEY = "xan"


class LoggingCommand:
    
    def __init__(self) -> None:
        user = settings.OPENOBSERVER_USER
        password = settings.OPENOBSERVER_PASSWORD
        self._bas64encoded_creds = base64.b64encode(bytes(user + ":" + password, "utf-8")).decode("utf-8")

        
    def send(self, stream: str, data: dict) -> requests.Response:
        headers = {"Content-type": "application/json", "Authorization": "Basic " + self._bas64encoded_creds}
        org = settings.OPENOBSERVER_ORGANIZATION
        stream = str(data[stream]).lower()
        openobserve_host = settings.OPENOBSERVER_HOST
        openobserve_url = openobserve_host + "/api/" + org + "/" + stream + "/_json"

        res = requests.post(openobserve_url, headers=headers, data=json.dumps(data))
        return res
       


import requests
from typing import Optional
from uuid import uuid4
from django.http import HttpRequest

from nkunyim_util.commands.authorize_command import AuthorizeCommand
from nkunyim_util.commands.delete_command import DeleteCommand
from nkunyim_util.commands.refresh_command import RefreshCommand
from nkunyim_util.commands.revoke_command import RevokeCommand
from nkunyim_util.commands.token_command import TokenCommand
from nkunyim_util.commands.userinfo_command import UserinfoCommand

from nkunyim_util.models.message_model import MessageModel, MessageLevel
from nkunyim_util.models.oauth2_model import OAuth2Model
from nkunyim_util.models.token_model import TokenModel
from nkunyim_util.models.userinfo_model import UserinfoModel

from .signals import token_data_updated


class OAuth2Service:
    
    def __init__(self, issuer_url: str, req: HttpRequest) -> None:
        self.req = req
        self.issuer_url = issuer_url.rstrip('/')
        self.model: Optional[OAuth2Model] = None
        self.messages: list[MessageModel] = []
        self.headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        
        # Session key
        self.key = self._key()
     
        # Session model
        self._retrieve_model()
    
    
    def _key(self) -> str:
        parts = self.req.get_host().lower().split('.')
        if len(parts) >= 2:
            domain = '.'.join(parts[-2:])
            subdomain = parts[-3] if len(parts) > 2 else "www"
            return f"{subdomain}.{domain}"
        return "www.localhost"
    
    
    def _save_model(self) -> None:
        if self.model:
            self.req.session[self.key] = self.model.model_dump()
            self.req.session.modified = True
          
            
    def _retrieve_model(self) -> None:
        if bool(self.key in self.req.session and self.req.session[self.key]):
            session = self.req.session[self.key]
            self.model = OAuth2Model(**session)
    
    
    def get_authorization_url(self, command: AuthorizeCommand) -> str:
        state = str(uuid4())
        nonce = uuid4().hex
        self.model = OAuth2Model(state=state, nonce=nonce, code="")
        self._save_model()
        
        authorize_url = (f"{self.issuer_url}/identity/authorize/?"
                         + f"response_type={command.response_type}&"
                         + f"client_id={command.client_id}&"
                         + f"redirect_uri={command.redirect_uri}&"
                         + f"scope={command.client_scope}&"
                         + f"state={state}&"
                         + f"nonce={nonce}"
        )
        
        return authorize_url
    
        
    def authorize_access_token(self, command: TokenCommand) -> Optional[TokenModel]:
        try:
            self._retrieve_model()
            if not self.model:
                self.messages.append(MessageModel(level=MessageLevel.WARNING, message="OAuth2 session data not found."))
                return None
            
            code = self.req.GET.get("code", "")
            state = self.req.GET.get("state", None)
            
            if not bool(state and self.model.state == state):
                self.messages.append(MessageModel(level=MessageLevel.ERROR, message="OAuth2 state mismatch."))
                return None
            
            error = self.req.GET.get("error", None)
            error_description = self.req.GET.get("error_description", None)
            if bool(error):
                self.messages.append(MessageModel(level=MessageLevel.ERROR, message=f"OAuth2 Error: [{error}] - {error_description}"))
                return None
            
            self.model.code = code
            self._save_model()
        
            # Set code in command
            command.code = code
            
            response = requests.post(url=self.issuer_url + "/identity/token/", data=command.model_dump(), headers=self.headers)
            if not response.ok:
                self.messages.append(MessageModel(level=MessageLevel.WARNING, message="Failed to retrieve token data from API."))
                return None
            
            json_data = response.json()
            token_data = TokenModel(**json_data)
            
            # Inform interested parties
            token_data_updated.send(sender=TokenModel, instance=token_data)
            
            return token_data
        except:
            self.messages.append(MessageModel(level=MessageLevel.ERROR, message="Failed to retrieve token data from API."))
            return None



    def refresh_access_token(self, access_token: str, command: RefreshCommand) -> Optional[TokenModel]:
        try:
            self.headers['Authorization'] = "JWT " + access_token
            response = requests.post(url=self.issuer_url + "/identity/refresh_token/", data=command.model_dump(), headers=self.headers)
            if not response.ok:
                self.messages.append(MessageModel(level=MessageLevel.WARNING, message="Failed to refresh token data from API."))
                return None
            
            json_data = response.json()
            token_data = TokenModel(**json_data)
            
            # Inform interested parties
            token_data_updated.send(sender=TokenModel, instance=token_data)
            
            return token_data
        except:
            self.messages.append(MessageModel(level=MessageLevel.ERROR, message="Failed to refresh token data from API."))
            return None
    


    def revoke_token(self, access_token: str, command: RevokeCommand) -> Optional[TokenModel]:
        try:
            self.headers['Authorization'] = "JWT " + access_token
            response = requests.post(url=self.issuer_url + "/identity/revoke_token/", data=command.model_dump(), headers=self.headers)
            if not response.ok:
                self.messages.append(MessageModel(level=MessageLevel.WARNING, message="Failed to revoke token data from API."))
                return None
            
            json_data = response.json()
            return TokenModel(**json_data)
        except:
            self.messages.append(MessageModel(level=MessageLevel.ERROR, message="Failed to revoke token data from API."))
            return None


    def introspect(self, access_token: str, command: RefreshCommand) -> Optional[UserinfoModel]:
        try:
            self.headers['Authorization'] = "JWT " + access_token
            response = requests.post(url=self.issuer_url + "/identity/introspect/", data=command.model_dump(), headers=self.headers)
            if not response.ok:
                self.messages.append(MessageModel(level=MessageLevel.WARNING, message="Failed to introspect user data from API."))
                return None
            
            json_data = response.json()
            return UserinfoModel(**json_data)
        except:
            self.messages.append(MessageModel(level=MessageLevel.ERROR, message="Failed to introspect user data from API."))
            return None
        

    def userinfo(self, access_token: str, command: UserinfoCommand) -> Optional[UserinfoModel]:
        try:
            self.headers['Authorization'] = "JWT " + access_token
            response = requests.post(url=self.issuer_url + "/identity/userinfo/", data=command.model_dump(), headers=self.headers)
            if not response.ok:
                self.messages.append(MessageModel(level=MessageLevel.WARNING, message="Failed to retrieve userinfo data from API."))
                return None
            
            json_data = response.json()
            return UserinfoModel(**json_data)
        except:
            self.messages.append(MessageModel(level=MessageLevel.ERROR, message="Failed to retrieve userinfo data from API."))
            return None


    def logout(self, access_token: str, command: DeleteCommand) -> Optional[TokenModel]:
        try:
            self.headers['Authorization'] = "JWT " + access_token
            response = requests.post(url=self.issuer_url + "/identity/logout/", data=command.model_dump(), headers=self.headers)
            if not response.ok:
                self.messages.append(MessageModel(level=MessageLevel.WARNING, message="Failed to logout token data from API."))
                return None
            
            json_data = response.json()
            return TokenModel(**json_data)
        except:
            self.messages.append(MessageModel(level=MessageLevel.ERROR, message="Failed to logout token data from API."))
            return None


from pydantic import BaseModel

from nkunyim_util.models.message_model import MessageModel


class ViewBag(BaseModel):
    ok: bool = False
    data: dict = {}
    messages: list[MessageModel] = []
    page_url: str = ""
    next_url: str = ""




from rest_framework import HTTP_HEADER_ENCODING, authentication

from nkunyim_util.services.user_service import UserService



class JWTAuthentication(authentication.BaseAuthentication):
    
    def __init__(self, *args, **kwargs):        
        super().__init__(*args, **kwargs)
                            
        self.authentication_headers = ["JWT", "Bearer", ]
        self.www_authenticate_realm = 'api' 


    def authenticate_header(self, request): # type: ignore
        return '{0} realm="{1}"'.format(
            self.authentication_headers[0],
            self.www_authenticate_realm,
        )
        
    def authenticate(self, request):
        header = self.get_header(request)
        if header is None:
            return None
                
        if isinstance(header, bytes):
            header = header.decode("utf-8") 
        
        token = self.get_raw_token(header)
        if token is None:
            return None
        
        user_service = UserService(req=request)
        user_model = user_service.api(token=token)
        if not user_model:
            return None
        
        # Return Auth Credentials
        return (user_model, None)


    def get_header(self, request):
        """
        Extracts the header containing the JSON web token from the given
        request.
        """
        header = request.META.get('HTTP_AUTHORIZATION', None)
        if header is None:
            return None

        if isinstance(header, str):
            # Work around django test client oddness
            header = header.encode(HTTP_HEADER_ENCODING)

        return header
    
    
    def get_raw_token(self, header):
        """
        Extracts an unvalidated JSON web token from the given "Authorization"
        header value.
        """
        parts = header.split()

        if len(parts) == 0:
            # Empty AUTHORIZATION header sent
            return None

        if parts[0] not in self.authentication_headers:
            # Assume the header does not contain a JSON web token
            return None
        
        if len(parts) != 2:
            return None

        return parts[1]
    
   
    


from enum import Enum
from pydantic import BaseModel


class MessageLevel(str, Enum):
    INFO = "info"
    SUCCESS =  "success"
    WARNING = "warning"
    ERROR =  "error"

        
class MessageModel(BaseModel):
    level: MessageLevel = MessageLevel.ERROR
    message: str




from typing import Optional
from pydantic import BaseModel


class UserAgentModel(BaseModel):
    is_mobile: bool
    is_tablet: bool
    is_touch_capable: bool
    is_pc: bool
    is_bot: bool
    is_email_client: bool
    browser_name: str
    browser_version: str
    os_name: str
    os_version: str
    device_name: str
    device_brand: Optional[str]
    device_model: Optional[str]
    



import base64
import logging
import requests
from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.encryption.rsa_encryption import RSAEncryption
from nkunyim_util.logging.logging_command import LOG_TYPE_API
from nkunyim_util.logging.logging_handler import LoggingHandler
from nkunyim_util.services.logging_service import LoggingService
from nkunyim_util.services.session_service import SessionService


# Get a logger instance
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# User handler
logging_handler = LoggingHandler()
logger.addHandler(logging_handler)

API_LOGGING_URL_KEY = "uri"
API_LOGGING_HEADERS_KEY = "headers"
API_LOGGING_BODY_KEY = "body"
API_LOGGING_METHOD_KEY = "method"
API_LOGGING_STATUS_KEY = "status"


class NkunyimApiClient:
    def __init__(self, req: HttpRequest, name: str) -> None:
        self.headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        self.req = req
        self.name = name.upper()
        self._init_base_url()
        self._init_headers()


    def _init_base_url(self) -> None:
        try:
            base_url = str(settings.NKUNYIM_SERVICES[self.name])
            self.base_url = base_url.rstrip('/')
        except KeyError:
            raise Exception(f"Service configuration '{self.name}' is not defined in settings.")


    def _init_headers(self) -> None:
        encryption = RSAEncryption()
        encrypted = encryption.encrypt(plain_text=self.name, name=self.name)
        token = base64.b64encode(encrypted).decode('utf-8')
        self.headers['Xan-Authorization'] = token

        session_service = SessionService(req=self.req)
        if session_service.has_lifetime:
            token_model = session_service.get_token()
            if token_model:
                self.headers['Authorization'] = f'JWT {token_model.access_token}'


    def _build_url(self, path: str) -> str:
        base = self.base_url
        if not base.endswith('/api'):
            base += '/api'

        if not path.startswith('/'):
            path = '/' + path

        return base + path
    
    
    def _make_log_extra(self, res: requests.Response) -> dict:
        service = LoggingService(req=self.req)
        model = service.get(xan=LOG_TYPE_API, service_name=self.name)
        return {
            **model.model_dump(),
            'request': {
                API_LOGGING_URL_KEY: res.request.url,
                API_LOGGING_METHOD_KEY: res.request.method,
                API_LOGGING_HEADERS_KEY: res.request.headers,
                API_LOGGING_BODY_KEY: res.request.body
            },
            'response': {
                API_LOGGING_STATUS_KEY: res.status_code,
                API_LOGGING_HEADERS_KEY: res.headers,
                API_LOGGING_BODY_KEY: res.text
            }
        }


    def get(self, path: str):
        response = requests.get(self._build_url(path), headers=self.headers)
        if not response.ok:
            extra = self._make_log_extra(res=response)
            logger.info(path, extra=extra)
        return response


    def post(self, path: str, data: dict):
        url = self._build_url(path)
        if not url.endswith('/'):
            url += '/'
        response = requests.post(url, data=data, headers=self.headers)
        if not response.ok:
            extra = self._make_log_extra(res=response)
            logger.info(path, extra=extra)
        return response


    def put(self, path: str, data: dict):
        response = requests.put(self._build_url(path), data=data, headers=self.headers)
        if not response.ok:
            extra = self._make_log_extra(res=response)
            logger.info(path, extra=extra)
        return response


    def patch(self, path: str, data: dict):
        response = requests.patch(self._build_url(path), data=data, headers=self.headers)
        if not response.ok:
            extra = self._make_log_extra(res=response)
            logger.info(path, extra=extra)
        return response


    def delete(self, path: str):
        response = requests.delete(self._build_url(path), headers=self.headers)
        if not response.ok:
            extra = self._make_log_extra(res=response)
            logger.info(path, extra=extra)
        return response




MODULE_PAGE_PERMISSIONS = [
    'create',
    'view',
    'edit',
    'delete',
]

MODULE_PAGE_PERMISSIONS_EXT = [
    'print',
    'import',
    'export',
    'upload', 
    'download',
    'activate',
    'deactivate',
]

MODULE_NODE_CHOICES = (
    ("dashboard", "Dashboard"),
    ("modules", "Modules"),
    ("account", "Account"),
    ("system", "System"),
)

SKIN_COLOUR_CHOICES = (
    ("behance", "Behance"),
    ("blue", "Blue"),
    ("css3", "CSS3"),
    ("dribbble", "Dribbble"),
    ("dropbox", "DropBox"),
    ("facebook", "Facebook"),
    ("flickr", "FlickR"),
    ("foursquare", "FourSquare"),
    ("github", "GitHub"),
    ("google-plus", "Google+"),
    ("green", "Green"),
    ("html5", "HTML5"),
    ("indigo", "Indigo"),
    ("instagram", "Instagram"),
    ("linkedin", "LinkedIn"),
    ("nkunyim", "Nkunyim"),
    ("openid", "OpenID"),
    ("orange", "Orange"),
    ("pink", "Pink"),
    ("pinterest", "Pinterest"),
    ("purple", "Purple"),
    ("red", "Red"),
    ("reddit", "Reddit"),
    ("spotify", "Spotify"),
    ("stack-overflow", "StackOverflow"),
    ("teal", "Teal"),
    ("tumblr", "Tumblr"),
    ("twitter", "Twitter"),
    ("vimeo", "Vimeo"),
    ("vine", "Vine"),
    ("vk", "Vk"),
    ("xing", "Xing"),
    ("yahoo", "Yahoo"),
    ("yellow", "Yellow"),
)

APPLICATION_MODE_CHOICES = (
    ("web", "Web"),
    ("mob", "Mobile"),
    ("api", "Service"),
    ("cli", "Console"),
)

PROJECT_LICENSE_CHOICES = (
    ("personal", "Personal"),
    ("family", "Family"),
    ("business", "Business"),
    ("public", "Public"),
)

ENTITY_TYPE_CHOICES = (
    ("personal", "Personal"),
    ("business", "Business"),
)

RESPONSE_TYPE_CHOICES = (
    ('code', 'Code'),
    ('token', 'Token'),
)

OTP_TYPE_CHOICES = (
    ('sms', 'SMS'),
    ('email', 'Email'),
    ('push', 'Push'),
    ('totp', 'TOTP'),
)

PERSON_TITLE_CHOICES = (
    ("Mr", "Mr."),
    ("Ms", "Ms."),
    ("Mrs", "Mrs."),
    ("Miss", "Miss."),
    ("Dr", "Dr."),
    ("Rev", "Rev."),
    ("Na", "NA"),
)

GENDER_CHOICES = (
    ("M", "Male"),
    ("F", "Female"),
    ("N", "NA"),
)

MARITAL_CHOICES = (
    ("S", "Single"),
    ("M", "Married"),
    ("D", "Divorced"),
    ("W", "Widow(er)"),
    ("O", "Other"),
    ("N", "NA"),
)

ADDRESS_NAME_CHOICES = [
    ('email', 'Email Address'),
    ('phone', 'Phone Number'),
    ('digital', 'Digital Address'),
    ('postal', 'Postal Address'),
    ('website', 'Website URL'),
    ('social', 'Social Media ID'),
    ('geoloc', 'Geo Location'),
    ('other', 'Other Contact'),
]

IDENTITY_NAME_CHOICES = [
    ('national', 'National ID'),
    ('passport', 'Passport'),
    ('driving', 'Driving License'),
    ('student', 'Student ID'),
    ('other', 'Other Identity'),
]



from pydantic import BaseModel
from typing import Optional

from nkunyim_util.models.user_model import UserModel
from nkunyim_util.models.token_model import TokenModel



class SessionModel(BaseModel):
    token: Optional[TokenModel]
    user: Optional[UserModel]